module Domains

import ..Geometry: Geometry, float_type
using IntervalSets
export RectangleDomain

"""
    AbstractDomain

A domain represents a region of space.
"""
abstract type AbstractDomain end

const BCTagType = Union{Nothing, Tuple{Symbol, Symbol}}

float_type(domain::AbstractDomain) = float_type(coordinate_type(domain))

"""
    boundary_names(obj::Union{AbstractDomain, AbstractMesh, AbstractTopology})

A tuple or vector of unique boundary names of a spatial domain.
"""
function boundary_names end

struct IntervalDomain{CT, B} <: AbstractDomain where {
    CT <: Geometry.Abstract1DPoint{FT},
    B <: BCTagType,
} where {FT}
    coord_min::CT
    coord_max::CT
    boundary_names::B
end

isperiodic(domain::IntervalDomain) = isnothing(domain.boundary_names)
boundary_names(domain::IntervalDomain) =
    isperiodic(domain) ? () : unique(domain.boundary_names)

"""
    IntervalDomain(coord⁻, coord⁺; periodic=true)
    IntervalDomain(coord⁻, coord⁺; boundary_names::Tuple{Symbol,Symbol})

Construct a `IntervalDomain`, the closed interval is given by `coord⁻`, `coord⁺` coordinate arguments.

Either a `periodic` or `boundary_names` keyword argument is required.
"""
function IntervalDomain(
    coord_min::Geometry.Abstract1DPoint,
    coord_max::Geometry.Abstract1DPoint;
    periodic = false,
    boundary_tags = nothing, # TODO: deprecate this
    boundary_names::BCTagType = boundary_tags,
)
    if !periodic && isnothing(boundary_names)
        throw(
            ArgumentError(
                "if `periodic=false` then an `boundary_names::Tuple{Symbol,Symbol}` keyword argument is required.",
            ),
        )
    end
    IntervalDomain(promote(coord_min, coord_max)..., boundary_names)
end
IntervalDomain(coords::ClosedInterval; kwargs...) =
    IntervalDomain(coords.left, coords.right; kwargs...)

coordinate_type(::IntervalDomain{CT}) where {CT} = CT
Base.eltype(domain::IntervalDomain) = coordinate_type(domain)

function Base.show(io::IO, domain::IntervalDomain)
    print(
        io,
        nameof(typeof(domain)),
        "($(domain.coord_min), $(domain.coord_max); ",
    )
    if isperiodic(domain)
        print(io, "periodic=true)")
    else
        print(io, "boundary_names = $(domain.boundary_names))")
    end
end

struct RectangleDomain{I1 <: IntervalDomain, I2 <: IntervalDomain} <:
       AbstractDomain
    interval1::I1
    interval2::I2
end
Base.:*(interval1::IntervalDomain, interval2::IntervalDomain) =
    RectangleDomain(interval1, interval2)

boundary_names(domain::RectangleDomain) = unique([
    boundary_names(domain.interval1)...,
    boundary_names(domain.interval2)...,
])

"""
    RectangleDomain(x1::ClosedInterval, x2::ClosedInterval;
        x1boundary::Tuple{Symbol,Symbol},
        x2boundary::Tuple{Symbol,Symbol},
        x1periodic = false,
        x2periodic = false,
    )

Construct a `RectangularDomain` in the horizontal.
If a given x1 or x2 boundary is not periodic, then `x1boundary` or `x2boundary` boundary name keyword arguments must be supplied.
"""
function RectangleDomain(
    x1::ClosedInterval{X1CT},
    x2::ClosedInterval{X2CT};
    x1periodic = false,
    x2periodic = false,
    x1boundary::BCTagType = nothing,
    x2boundary::BCTagType = nothing,
) where {X1CT <: Geometry.Abstract1DPoint, X2CT <: Geometry.Abstract1DPoint}
    interval1 =
        IntervalDomain(x1; periodic = x1periodic, boundary_names = x1boundary)
    interval2 =
        IntervalDomain(x2; periodic = x2periodic, boundary_names = x2boundary)
    return interval1 * interval2
end


function Base.show(io::IO, domain::RectangleDomain)
    indent = get(io, :indent, 0)
    println(io, nameof(typeof(domain)), "(")
    println(io, " "^(indent + 2), domain.interval1)
    println(io, " "^(indent + 2), domain.interval2)
    print(io, " "^indent, ")")
end
coordinate_type(domain::RectangleDomain) = typeof(
    Geometry.product_coordinates(
        domain.interval1.coord_min,
        domain.interval2.coord_min,
    ),
)

"""
    SphereDomain(radius)

A domain representing the surface of a sphere with radius `radius`.
"""
struct SphereDomain{FT} <: AbstractDomain where {FT <: AbstractFloat}
    radius::FT
end
Base.show(io::IO, domain::SphereDomain) =
    print(io, nameof(typeof(domain)), "($(domain.radius))")

boundary_names(::SphereDomain) = ()
coordinate_type(::SphereDomain{FT}) where {FT} = Geometry.Cartesian123Point{FT}

end # module
